
from ._main import plot
from ._main import pie
from ._main import bar_chart
from ._main import regplot
from ._main import imshow
from ._main import hist
from ._main import scatter
from ._main import contour
from ._main import dumbbell_plot
from ._main import ridge
from ._main import parallel_coordinates
from ._taylor_plot import taylor_plot

from .utils import process_axis


__version__ = "0.19"
